
from twisted.internet import defer, reactor
from twisted.words.protocols.jabber.client import IQ

from model import Activity
from test_component import TestCase
from util import elem_iq, elem, xpath_query, elem_find_children

import ns

def parse_properties(node, ns):
    properties = {}

    for child in elem_find_children(node, ns, 'properties'):
        for property in elem_find_children(child, ns, 'property'):
            name = property['name']
            type = property['type']
            value = property.children[0]
            properties[name] = (type, value)

    return properties

def parse_activities_node(node):
    activities = []

    for child in node.elements():
        if child.name == 'activity':
            id = child.getAttribute('id')
            room = child.getAttribute('room')
            assert id
            assert room
            props = parse_properties(child, ns.ACTIVITY_PROP)
            activities.append(Activity(id, room, props))

            for child in child.elements():
                if child.name == 'buddy':
                    jid = child.getAttribute('jid')
                    assert jid
                    activities[-1].members.add(jid)

    return sorted(activities)

def parse_activities_result(stanza):
    assert stanza.getAttribute('type') == 'result'
    query = stanza.firstChildElement()
    assert query.name in ['query', 'view']

    return parse_activities_node(query)

class ActivityIqTest(TestCase):
    """Activity query with no results."""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        iq = elem_iq(self.server, 'get',
            from_='bob@foo.org', to='gadget.bar.org')(
                elem(ns.ACTIVITY, 'query'))
        iq.send()

        return self.done

    def iq(self, stanza):
        assert parse_activities_result(stanza) == []
        self.done.callback(None)

class ActivityIqTest2(TestCase):
    """Activity query with two results."""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_activity(1, 'foo', {})
        model.create_activity(2, 'bar', {})

        iq = elem_iq(self.server, 'get',
            from_='bob@foo.org', to='gadget.bar.org')(
                elem(ns.ACTIVITY, 'query'))
        iq.send()

        return self.done

    def iq(self, stanza):
        activities = parse_activities_result(stanza)
        assert activities == [Activity('1', 'foo'), Activity('2', 'bar')]
        self.done.callback(None)

class ActivityIqTest3(TestCase):
    """Activity query with a result which has members."""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        activity = model.create_activity(1, 'foo', {})
        activity.members.add('bob@foo.org')
        activity.members.add('tom@foo.org')

        iq = elem_iq(self.server, 'get',
            from_='bob@foo.org', to='gadget.bar.org')(
                elem(ns.ACTIVITY, 'query'))
        iq.send()

        return self.done

    def iq(self, stanza):
        activities = parse_activities_result(stanza)
        assert len(activities) == 1
        assert activities[0].id == '1'
        assert activities[0].members == set(['bob@foo.org', 'tom@foo.org'])
        self.done.callback(None)

class ActivityIqTest4(TestCase):
    """Activity query using properties with two results."""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_activity(1, 'foo', {'type': ('str', 'Paint')})
        model.create_activity(2, 'bar', {'type': ('str', 'Connect')})
        model.create_activity(3, 'test', {'type': ('str', 'Paint')})

        iq = elem_iq(self.server, 'get',
                from_='bob@foo.org', to='gadget.bar.org')(
            elem(ns.ACTIVITY, 'query')(
                elem('activity')(
                    elem(ns.ACTIVITY_PROP, 'properties')(
                        elem('property', type='str', name='type')(u'Paint')))))
        iq.send()

        return self.done

    def iq(self, stanza):
        activities = parse_activities_result(stanza)
        props = {'type': ('str', 'Paint')}
        assert activities == [Activity('1', 'foo', props),
                Activity('3', 'test', props)]
        self.done.callback(None)

class ActivityIqTest5(TestCase):
    """Activity query using muc with one result."""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_activity(1, 'foo', {'type': ('str', 'Paint')})
        model.create_activity(2, 'bar', {'type': ('str', 'Connect')})
        model.create_activity(3, 'test', {'type': ('str', 'Paint')})

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.ACTIVITY, 'query'))
        activity = query.addElement((None, 'activity'))
        activity['room'] = 'bar'
        iq.send()

        return self.done

    def iq(self, stanza):
        activities = parse_activities_result(stanza)
        assert activities == [Activity('2', 'bar', {'type': ('str', 'Connect')})]
        self.done.callback(None)

class ActivityIqTest6(TestCase):
    """Activity query using participant with two results."""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        activity = model.create_activity(1, 'foo', {'type': ('str', 'Paint')})
        activity.members.add('alice@foo.org')
        activity.members.add('bob@foo.org')
        activity = model.create_activity(2, 'bar',
            {'type': ('str', 'Connect')})
        activity.members.add('alice@foo.org')
        activity = model.create_activity(3, 'test', {'type': ('str', 'Paint')})
        activity.members.add('bob@foo.org')

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.ACTIVITY, 'query'))
        activity = query.addElement((None, 'activity'))
        buddy = activity.addElement((None, 'buddy'))
        buddy['jid'] = 'alice@foo.org'
        iq.send()

        return self.done

    def iq(self, stanza):
        activities = parse_activities_result(stanza)
        assert len(activities) == 2
        assert activities[0].id == '1'
        assert activities[1].id == '2'
        self.done.callback(None)

class ActivityIqTest7(TestCase):
    """Activity query using participants and properties with two results."""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        activity = model.create_activity(1, 'foo', {'type': ('str', 'Paint')})
        activity.members.add('alice@foo.org')
        activity.members.add('bob@foo.org')
        activity = model.create_activity(2, 'bar',
            {'type': ('str', 'Connect')})
        activity.members.add('alice@foo.org')
        activity = model.create_activity(3, 'test', {'type': ('str', 'Paint')})
        activity.members.add('bob@foo.org')

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.ACTIVITY, 'query'))
        activity = query.addElement((None, 'activity'))
        buddy = activity.addElement((None, 'buddy'))
        buddy['jid'] = 'alice@foo.org'
        properties = activity.addElement((ns.ACTIVITY_PROP, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'str'
        property['name'] = 'type'
        property.addContent('Paint')
        iq.send()

        return self.done

    def iq(self, stanza):
        activities = parse_activities_result(stanza)
        assert len(activities) == 1
        assert activities[0].id == '1'
        self.done.callback(None)

class ActivityIqInvalidPropertyTest(TestCase):
    """Activity query with an invalid property"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.ACTIVITY, 'query'))
        activity = query.addElement((None, 'activity'))
        properties = activity.addElement((ns.ACTIVITY_PROP, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'wtf'
        property['name'] = 'nothing'
        property.addContent('hey')
        iq.send()

        return self.done

    def iq(self, stanza):
        assert stanza.getAttribute('type') == 'error'
        error = stanza.firstChildElement()
        assert error.name == 'error'
        e = error.firstChildElement()
        assert e.name == 'invalid-property'
        assert e.uri == ns.ACTIVITY_PROP

        self.done.callback(None)

class ActivityViewByProperties(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        activity = model.create_activity(1, 'foo', {})
        activity.members.add('charles@localhost')
        model.create_activity(2, 'bar', {'color': ('str', 'blue')})

        model.create_buddy('charles@localhost', {'color': ('str', 'red')})
        model.create_buddy('georges@localhost', {'color': ('str', 'green')})

        iq = elem_iq(self.server, 'get',
                from_='bob@foo.org/Resource', to='gadget@bar.org')(
            elem(ns.ACTIVITY, 'view', id='query1', size='20')(
                elem('activity')(
                    elem(ns.ACTIVITY_PROP, 'properties')(
                        elem('property', type='str', name='color')(u'blue')))))
        iq.send()

        return self.done

    def iq(self, stanza):
        views = self.service.activity_views['bob@foo.org']
        view = views['query1']
        assert view.size == 20

        activities = parse_activities_result(stanza)
        assert len(activities) == 1

        self.service.model.activity_update(1, {'color': ('str', 'blue')})

        self.server.addOnetimeObserver('//message', self.message1)

    def message1(self, stanza):
        self.assertEquals('bob@foo.org', stanza['to'])
        #self.assertEquals('gadget@bar.org', stanza['from'])
        added = xpath_query('/message/added[@xmlns="%s"]'
                % ns.ACTIVITY, stanza)[0]
        assert added['id'] == 'query1'
        activity = added.firstChildElement()
        assert activity.name == 'activity'
        assert activity['id'] == '1'
        assert activity['room'] == 'foo'

        activities = parse_activities_node(added)
        assert activities == [Activity('1', 'foo', {'color': ('str', 'blue')},
            set(['charles@localhost']))]

        buddy_node = xpath_query('/added/activity/buddy', added)[0]
        props = parse_properties(buddy_node, ns.BUDDY_PROP)
        assert props == {'color': ('str', 'red')}

        # add a member to activity
        self.server.addOnetimeObserver('//message', self.message2)
        self.service.model.activity_add_member(1, 'georges@localhost')

    def message2(self, stanza):
        activity = xpath_query('/message/activity[@xmlns="%s"]'
                % ns.ACTIVITY, stanza)[0]
        assert activity['id'] == '1'
        assert activity['room'] == 'foo'
        assert activity['view'] == 'query1'
        joined = activity.firstChildElement()
        assert joined.name == 'joined'
        assert joined['jid'] == 'georges@localhost'

        props = parse_properties(joined, ns.BUDDY_PROP)
        assert props == {'color': ('str', 'green')}

        # member changed his properties
        self.server.addOnetimeObserver('//message', self.message3)
        self.service.model.buddy_update('georges@localhost', {'color': ('str', 'blue')})

    def message3(self, stanza):
        change = xpath_query('/message/change[@xmlns="%s"]'
                % ns.BUDDY, stanza)[0]
        assert change['jid'] == 'georges@localhost'
        assert change['id'] == 'query1'

        props = parse_properties(change, ns.BUDDY_PROP)
        self.assertEquals(props, {'color': ('str', 'blue')})

        # member changed his current-activity
        self.server.addOnetimeObserver('//message', self.message4)
        self.service.model.buddy_change_current_activity('georges@localhost',
            Activity('room1', 'id1', {}))

    def message4(self, stanza):
        change = xpath_query('/message/change[@xmlns="%s"]'
                % ns.BUDDY, stanza)[0]
        assert change['jid'] == 'georges@localhost'
        assert change['id'] == 'query1'
        activity = xpath_query('/change/activity[@xmlns="%s"]'
                % ns.CURRENT_ACTIVITY, change)[0]
        activity['room'] == 'room1'
        activity['id'] == 'id1'

        # remove a member from the activity
        self.server.addOnetimeObserver('//message', self.message5)
        self.service.model.activity_remove_member(1, 'georges@localhost')

    def message5(self, stanza):
        activity = xpath_query('/message/activity[@xmlns="%s"]'
                % ns.ACTIVITY, stanza)[0]
        assert activity['id'] == '1'
        assert activity['room'] == 'foo'
        assert activity['view'] == 'query1'
        left = activity.firstChildElement()
        assert left.name == 'left'
        assert left['jid'] == 'georges@localhost'

        # remove activity
        self.server.addOnetimeObserver('//message', self.message6)
        self.service.model.activity_remove(1)

    def message6(self, stanza):
        self.assertEquals('bob@foo.org', stanza['to'])
        #self.assertEquals('gadget@bar.org', stanza['from'])
        removed = xpath_query('/message/removed[@xmlns="%s"]'
                % ns.ACTIVITY, stanza)[0]
        assert removed['id'] == 'query1'
        activity = removed.firstChildElement()
        assert activity.name == 'activity'
        assert activity['id'] == '1'
        assert activity['room'] == 'foo'

        self.done.callback(None)

class ActivityIqRandomQuery(TestCase):
    """activity random view query"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_activity(1, 'foo', {})
        model.create_activity(2, 'bar', {})
        model.create_activity(3, 'barbar', {})

        iq = elem_iq(self.server, 'get',
            from_='bob@foo.org', to='gadget@bar.org')(
                elem(ns.ACTIVITY, 'view', id='query1', size='2'))
        iq.send()

        return self.done

    def iq(self, stanza):
        views = self.service.activity_views['bob@foo.org']
        view = views['query1']
        assert view.size == 2

        activities = parse_activities_result(stanza)
        assert len(activities) == 2

        # close the view
        message = elem('message', from_='bob@foo.org/Telepathy', to='gadget@bar.org')(
            elem('close', xmlns=ns.ACTIVITY, id='query1'))

        self.server.send(message)
        reactor.callLater(0, self.later)

    def later(self):
        assert self.service.activity_views == {}
        # view was removed from the model too
        assert len(self.model.activity_views) == 0
        self.done.callback(None)

class ActivityIqViewNoId(TestCase):
    """activity random view query without ID"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        view = iq.addElement((ns.ACTIVITY, 'view'))
        view['size'] = '3'
        iq.send()

        return self.done

    def iq(self, stanza):
        assert stanza.getAttribute('type') == 'error'
        assert len(xpath_query('/iq/error/invalid-property[@xmlns="%s"]' %
            ns.ACTIVITY, stanza)) == 1

        self.done.callback(None)

class ActivityIqViewExist(TestCase):
    """activity random view query using an existing ID"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        # create a fake view with 'query1' as ID
        self.service.activity_views['bob@foo.org'] = {}
        self.service.activity_views['bob@foo.org']['query1'] = True

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        view = iq.addElement((ns.ACTIVITY, 'view'))
        view['id'] = 'query1'
        view['size'] = '3'
        iq.send()

        return self.done

    def iq(self, stanza):
        assert stanza.getAttribute('type') == 'error'
        assert len(xpath_query('/iq/error/already-exists[@xmlns="%s"]' %
            ns.ACTIVITY, stanza)) == 1

        self.done.callback(None)

class ActivityNotInViewRemoved(TestCase):
    """activity random view query"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_activity(1, 'foo', {})
        model.create_activity(2, 'bar', {})

        iq = elem_iq(self.server, 'get',
            from_='bob@foo.org', to='gadget@bar.org')(
                elem(ns.ACTIVITY, 'view', id='query1', size='1'))
        iq.send()

        return self.done

    def iq(self, stanza):
        activities = parse_activities_result(stanza)
        activity = activities[0]

        # remove the activity which is not in the view
        if activity.id == '1':
            self.model.activity_remove(2)
        elif activity.id == '2':
            self.model.activity_remove(1)

        self.done.callback(None)

class ActivityViewByParticipants(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        # Alice is in activity 1 and 2
        # Charles is in activity 1 and 3
        activity = model.create_activity(1, 'foo', {})
        activity.members.add('alice@foo.org')
        activity.members.add('charles@foo.org')
        activity = model.create_activity(2, 'bar', {})
        activity.members.add('alice@foo.org')
        activity = model.create_activity(3, 'test', {})
        activity.members.add('charles@foo.org')

        iq = elem_iq(self.server, 'get',
                from_='bob@foo.org/Resource', to='gadget@bar.org')(
            elem(ns.ACTIVITY, 'view', id='query1', size='20')(
                elem('activity')(
                    elem('buddy', jid='alice@foo.org'),
                    elem('buddy', jid='charles@foo.org'))))
        iq.send()

        return self.done

    def iq(self, stanza):
        assert 'bob@foo.org' in self.service.activity_views
        activities = parse_activities_result(stanza)
        assert activities == [Activity('1', 'foo', {},
            set(['alice@foo.org', 'charles@foo.org']))]

        self.server.addOnetimeObserver('//message', self.message1)
        # Charles joined activity 2
        self.service.model.activity_add_member(2, 'charles@foo.org')

    def message1(self, stanza):
        self.assertEquals('bob@foo.org', stanza['to'])
        added = xpath_query('/message/added[@xmlns="%s"]'
                % ns.ACTIVITY, stanza)[0]
        assert added['id'] == 'query1'
        activity = added.firstChildElement()

        activities = parse_activities_node(added)
        self.assertEquals(activities, [Activity('2', 'bar', {},
            set(['alice@foo.org', 'charles@foo.org']))])

        # Alice left activity 1
        self.server.addOnetimeObserver('//message', self.message2)
        self.service.model.activity_remove_member(1, 'alice@foo.org')

    def message2(self, stanza):
        self.assertEquals('bob@foo.org', stanza['to'])
        removed = xpath_query('/message/removed[@xmlns="%s"]'
                % ns.ACTIVITY, stanza)[0]
        assert removed['id'] == 'query1'
        activity = removed.firstChildElement()

        activities = parse_activities_node(removed)
        self.assertEquals(activities, [Activity('1', 'foo')])

        self.done.callback(None)

class ActivityViewByParticipantsAndProperties(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        activity1 = model.create_activity(1, 'room1@conf.foo.org',
            {'type': ('str', 'Connect')})
        activity1.members.add('alice@localhost')
        activity2 = model.create_activity(2, 'room2@conf.foo.org', {})
        activity2.members.add('alice@localhost')
        activity3 = model.create_activity(3, 'room3@conf.foo.org',
            {'type': ('str', 'Connect')})
        activity3.members.add('bob@localhost')

        # request Connect activities in which Alice is participanting
        iq = elem_iq(self.server, 'get',
                from_='bob@foo.org/Resource', to='gadget@bar.org')(
            elem(ns.ACTIVITY, 'view', id='query1', size='20')(
                elem('activity')(
                    elem('buddy', jid='alice@localhost'),
                    elem(ns.ACTIVITY_PROP, 'properties')(
                        elem('property', type='str', name='type')(u'Connect')))))
        iq.send()

        return self.done

    def iq(self, stanza):
        assert 'bob@foo.org' in self.service.activity_views
        activities = parse_activities_result(stanza)
        assert activities == [Activity('1', 'room1@conf.foo.org',
            {'type': ('str', 'Connect')}, set(['alice@localhost']))]

        self.server.addOnetimeObserver('//message', self.message1)
        # Alice joined activity3
        self.service.model.activity_add_member(3, 'alice@localhost')

    def message1(self, stanza):
        self.assertEquals('bob@foo.org', stanza['to'])
        added = xpath_query('/message/added[@xmlns="%s"]'
                % ns.ACTIVITY, stanza)[0]
        assert added['id'] == 'query1'
        activity = added.firstChildElement()

        activities = parse_activities_node(added)
        assert activities == [Activity('3', 'room3@conf.foo.org',
            {'type': ('str', 'Connect')},
            set(['alice@localhost', 'bob@localhost']))]

        self.server.addOnetimeObserver('//message', self.message2)
        # activity2 became a Connect activity
        self.service.model.activity_update(2, {'type': ('str', 'Connect')})

    def message2(self, stanza):
        self.assertEquals('bob@foo.org', stanza['to'])
        added = xpath_query('/message/added[@xmlns="%s"]'
                % ns.ACTIVITY, stanza)[0]
        assert added['id'] == 'query1'
        activity = added.firstChildElement()

        activities = parse_activities_node(added)
        assert activities == [Activity('2', 'room2@conf.foo.org',
            {'type': ('str', 'Connect')}, set(['alice@localhost']))]

        self.server.addOnetimeObserver('//message', self.message3)
        # Alice left activity1
        self.service.model.activity_remove_member(1, 'alice@localhost')

    def message3(self, stanza):
        self.assertEquals('bob@foo.org', stanza['to'])
        removed = xpath_query('/message/removed[@xmlns="%s"]'
                % ns.ACTIVITY, stanza)[0]
        assert removed['id'] == 'query1'
        activity = removed.firstChildElement()
        assert activity.name == 'activity'
        assert activity['id'] == '1'
        assert activity['room'] == 'room1@conf.foo.org'

        self.done.callback(None)

class ActivityIqViewInvalidSize(TestCase):
    """activity view query with an invalid size attribute"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        view = iq.addElement((ns.ACTIVITY, 'view'))
        view['id'] = 'query1'
        view['id'] = 'foo'
        iq.send()

        return self.done

    def iq(self, stanza):
        assert stanza.getAttribute('type') == 'error'
        error = stanza.firstChildElement()
        assert error.name == 'error'
        e = error.firstChildElement()
        assert e.name == 'invalid-property'

        self.done.callback(None)
